from .variables import *
import speech_recognition as _SR
import whisper as _whisper
from vosk import Model as _vosk_model, KaldiRecognizer as _kaldi
import pyaudio  
from pydub import AudioSegment
import requests
import base64
import numpy as np

Microphone= _SR.Microphone
_R = _SR.Recognizer()

class Vosk:
    def __init__(self,model_language:str=Languages.English,model_size:str=VoskModelSize.Small):
        self._model=VoskModels.GetModel(model_language,model_size)
        self.options={'kaldi_h':16000}
    def install_model(self):
        return VMM.install_model(self._model,prints=True)
    def init(self):
        if self.install_model():
            self.recognizer = _kaldi(_vosk_model(self.load_model()),self.options['kaldi_h'])
            return True
        return False
    def load_model(self):
        return VMM.get_model_path(self._model)
    def set_model(self,model:dict):
        self._model = model
    def get_model(self):
        return self.recognizer
    def __call__(self):
        return self.get_model()


class Whisper:
    def __init__(self,whispermodel:str=WhisperModels.Base):
        self._whispermodel = whispermodel
    def load_model(self):
        self.recognizer = _whisper.load_model(self._whispermodel)
    def init(self):
        self.load_model()
        return True
    def get_model(self):
        return self.recognizer
    def __call__(self):
        return self.get_model()

class Wit:
    def __init__(self,apikey:str,apikey_language:str):
        self.apikey=apikey
        self.lang=apikey_language
    def recognize(self,audio_data):
        return _R.recognize_wit(audio_data,self.apikey)

class VoskAPI:
    def __init__(self,language:str='en'):
        self.language = language
    def recognize(self,audio_data):
        return _R.recognize_vosk(audio_data,self.language)

class Google:
    def __init__(self,language:str='en-US'):
        self.language=language
    def recognize(self,audio_data):
        return _R.recognize_google(audio_data,language=self.language)

class Sphinx:
    def __init__(self,language='en-US'):
        self.language= language
    def recognize(self,audio_data):
        return _R.recognize_sphinx(audio_data,self.language)

class CloudFlare:
    def __init__(self,account_id:str,api_key:str):
        self.accid=account_id
        self.apikey = api_key
        self.headers={"Authorization":f"Bearer {self.apikey}"}
    def recognize_whisper(self,audio_data,language):
        out=requests.post(f"https://api.cloudflare.com/client/v4/accounts/{self.accid}/ai/run/@cf/openai/whisper-large-v3-turbo",headers=self.headers,json={"audio":base64.b64encode(audio_data).decode(),"language":language})
        out=out.json() if out.status_code == 200 else False
        return False if not out else out['result']['text']
    def recognize(self,audio_data):
        out=requests.post(f"https://api.cloudflare.com/client/v4/accounts/{self.accid}/ai/run/@cf/openai/whisper",headers=self.headers,json={"audio":list(audio_data)})
        return out

def convert_raw_to_audio_data(raw_audio: bytes, sample_rate: int, sample_width: int = 2,channels:int=1) -> _SR.AudioData:
    """
    Convert raw PCM audio data to speech_recognition.AudioData.

    Parameters:
        raw_audio (bytes): The raw PCM audio data.
        sample_rate (int): The sample rate of the audio.
        sample_width (int): The sample width in bytes (default is 2 for 16-bit audio).

    Returns:
        sr.AudioData: AudioData object compatible with speech_recognition.
    """

    # Convert raw audio to AudioData
    return _SR.AudioData(raw_audio, sample_rate, sample_width * channels)

def get_raw_audio(file_path: str, sample_rate: int = 16000) -> bytes:
    """
    Reads an audio file and returns it as raw audio data that can be used with PyAudio.

    Parameters:
        file_path (str): The path to the audio file.
        sample_rate (int): The desired sample rate for the audio.

    Returns:
        bytes: Raw PCM audio data that can be used with PyAudio.
    """
    try:
        # Load audio file with pydub
        audio = AudioSegment.from_file(file_path)

        # Convert to mono and set sample rate
        audio = audio.set_channels(1).set_frame_rate(sample_rate).set_sample_width(2)

        # Convert to raw data (PCM format)
        raw_data = audio.raw_data
        return raw_data

    except Exception as e:
        print(f"Error while processing the audio file: {e}")
        return b""

def play_raw_audio(raw_audio_data: bytes, sample_rate: int = 16000, channels: int = 1, sample_width: int = 2):
    """
    Plays raw audio data using PyAudio.

    Parameters:
        raw_audio_data (bytes): The raw PCM audio data to be played.
        sample_rate (int): The sample rate of the audio (default 16000 Hz).
        channels (int): Number of channels (1 for mono, 2 for stereo, etc., default is mono).
        sample_width (int): Sample width in bytes (default is 2 for 16-bit audio).
    """
    try:
        # Initialize PyAudio instance
        p = pyaudio.PyAudio()

        # Open a stream for playback
        stream = p.open(
            format=pyaudio.paInt16,  # 16-bit PCM format
            channels=channels,       # Mono or stereo
            rate=sample_rate,        # Sample rate (e.g., 16000)
            output=True              # Output stream
        )

        # Play the raw audio data in chunks
        stream.write(raw_audio_data)

        # Stop and close the stream
        stream.stop_stream()
        stream.close()

        # Terminate PyAudio
        p.terminate()

        print("Audio playback finished.")
    except Exception as e:
        print(f"Error during audio playback: {e}")




# These packages are imported from original outdated OMEGAPy-Project omg package lib
from .listtools import listing as LT
import math


class UnderSquareRoot:
    def __init__(self,y,x=2):
        self.y = y
        self.nus = y
        self.x = x
        if x == 0:
            raise AttributeError('Attribute x cannot be 0')
        self.answer = y**(1/x)
    def ret(self):
        return self.answer
    def __repr__(self):
        return f'{self.answer}'
def sqrt(y,x=2):
    return UnderSquareRoot(y,x).ret()

class K(int):
    def __init__(self, i):
        self.Integer = i
        super(K, self).__init__()
    def __eq__(self,x):
        return self.Integer == x
    def __ne__(self,x):
        return self.Integer != x
    def __repr__(self):
        if self.Integer < 999_999:
            return f"{self.Integer/1_000}K"
        elif self.Integer < 999_999_999:
            return f"{self.Integer/1_000_000}M"
        elif self.Integer < 999_999_999_999:
            return f"{self.Integer/1_000_000_000}B"
        elif self.Integer < 999_999_999_999_999:
            return f"{self.Integer/1_000_000_000_000}T"
        elif self.Integer < 999_999_999_999_999_999:
            return f"{self.Integer/1_000_000_000_000_000}q"
        else:
            return f"{self.Integer/1_000_000_000_000_000}q"       

pi = math.pi


usqrt = UnderSquareRoot



def N(number):
    if number >= 1:
        return number
    else:
        return float("nan")
        

def W(number):
    if number >= 0:
        return number
    else:
        return float("nan")

def Z(number):
    if int(number) == number or type(number) == int:
        return number
    else:
        return float("nan")

def Q(number):
    if type(number) != UnderSquareRoot:
        if type(number) == float:
            if len(str(number).split('.')[-1])>5:
                return float("nan")
        return number
    else:
        if Z(number.answer):
            return number
        else:
            return float("nan")


def Qp(number):
    if type(number) == UnderSquareRoot:
        if len(str(number.answer).split('.')[-1])>5:
            return number
        else:
            return float("nan")
    elif type(number) == float:
        return number if len(str(number).split('.')[-1])>5 else float("nan")
    else:
        return float("nan")


def qzz(vi):
    return vi**(math.pi/-vi)
    



def rotaten(n,sub):
    # return's -(n*sub)*2
    return -(abs(n))*(sub/0.5)

def rotatin(*ns,sub):
    # rotatin(3,3,sub=3)
    #
    # 3*3*4+(3*4)
    #
    # ns1*ns2*sub+(lenOfInput,sub)   # LenOfInput = n(3,3,4) = 3
    return -(abs(sum(ns)))*(sub/0.5)
    
def Mn(arr):
    y = 0
    for X in arr:
        y=y+1
    return y
 

def m_(x1,y1)->'y2':
    x2 = 100
    y2 = (y1*x2)/x1
    return y2

def m_p(x1,x2,y1)->'y2':
    y2 = (y1*x2)/x1
    return y2

def fx12y3_00(x1,x2)->'y3':
    y1=x1-x2
    y2=x2/x1
    x3=y1*y2
    y3=(x1**x2)-(y1**y2)+x3
    return y3




from ..tools.dicttools import perfectdict
from pathlib import Path
import random as _random
from pySmartDL import SmartDL as _download
from zipfile import ZipFile as _ZipFile
import shutil as _shutil
from whisper import available_models as _wa

class Languages:
    English='en';French='fr';Spanish='es';Mandarin='zh';Hindi='hi';Arabic='ar';Bengali='bn';Portuguese='pt';Russian='ru';Japanese='ja';Punjabi='pa';German='de';Korean='ko';Vietnamese='vi';Turkish='tr';Italian='it';Tamil='ta';Urdu='ur';Persian='fa';Dutch='nl';Thai='th';Malay='ms';Swahili='sw';Greek='el';Hungarian='hu';Polish='pl';Ukrainian='uk';Hebrew='he';Finnish='fi';Danish='da';Czech='cs';Slovak='sk';Norwegian='no';Swedish='sv';Romanian='ro';Bulgarian='bg';Croatian='hr';Serbian='sr';Slovenian='sl';Estonian='et';Latvian='lv';Lithuanian='lt';Icelandic='is';Irish='ga';Maltese='mt';Basque='eu';Catalan='ca';Galician='gl';Kurdish='ku';Amharic='am';Zulu='zu';Xhosa='xh';Yoruba='yo';Hausa='ha';Igbo='ig';Somali='so';Albanian='sq';Bosnian='bs';Macedonian='mk';Azerbaijani='az';Armenian='hy';Georgian='ka';Sinhala='si';Khmer='km';Lao='lo';Burmese='my';Mongolian='mn';Pashto='ps';Kazakh='kk';Uzbek='uz';Tajik='tg';Kyrgyz='ky';Turkmen='tk';Malayalam='ml';Kannada='kn';Telugu='te';Marathi='mr';Gujarati='gu';Oriya='or';Assamese='as';Nepali='ne';Tibetan='bo';Maori='mi';Samoan='sm';Tongan='to';Fijian='fj';Malagasy='mg';Sesotho='st';Tswana='tn';Shona='sn';Afrikaans='af';Lingala='ln';Luo='luo';Kinyarwanda='rw';Kirundi='rn';Quechua='qu';Aymara='ay';Guarani='gn';Nahuatl='nah'


Path().home().joinpath(".omgl/").joinpath("vosk_models/").mkdir(parents=True,exist_ok=True)

class VoskModelSize:
    Small = 'small'
    Medium = 'medium'
    Large = 'large'

class VoskModels:
    supported_languages = { # The data may not be accurate
    'en':{'small':{'name':'vosk-model-small-en-us-0.15','link':'https://alphacephei.com/vosk/models/vosk-model-small-en-us-0.15.zip'},'medium':{'name':'vosk-model-en-us-0.22','link':'https://alphacephei.com/vosk/models/vosk-model-en-us-0.22.zip'},'large':{'name':'vosk-model-en-us-0.42-gigaspeech','link':'https://alphacephei.com/vosk/models/vosk-model-en-us-0.42-gigaspeech.zip'}} # English
    ,'en_in':{'small':{'name':'vosk-model-small-en-in-0.4','link':'https://alphacephei.com/vosk/models/vosk-model-small-en-in-0.4.zip'},'medium':{'name':'vosk-model-en-in-0.5','link':'https://alphacephei.com/vosk/models/vosk-model-en-in-0.5.zip'}} # Indian English
    ,'cn':{'small':{'name':'vosk-model-small-cn-0.22','link':'https://alphacephei.com/vosk/models/vosk-model-small-cn-0.22.zip'},'medium':{'name':'vosk-model-cn-0.22','link':'https://alphacephei.com/vosk/models/vosk-model-cn-0.22.zip'},'large':{'name':'vosk-model-cn-kaldi-multicn-0.15','link':'https://alphacephei.com/vosk/models/vosk-model-cn-kaldi-multicn-0.15.zip'}} # Chinese
    ,'ru':{'small':{'name':'vosk-model-small-ru-0.22','link':'https://alphacephei.com/vosk/models/vosk-model-small-ru-0.22.zip'},'medium':{'name':'vosk-model-ru-0.42','link':'https://alphacephei.com/vosk/models/vosk-model-ru-0.42.zip'},'large':{'name':'vosk-model-ru-0.10','link':'https://alphacephei.com/vosk/models/vosk-model-ru-0.10.zip'}} # Russian
    ,'fr':{'small':{'name':'vosk-model-small-fr-0.22','link':'https://alphacephei.com/vosk/models/vosk-model-small-fr-0.22.zip'},'medium':{'name':'vosk-model-fr-0.22','link':'https://alphacephei.com/vosk/models/vosk-model-fr-0.22.zip'},'large':{'name':'vosk-model-fr-0.6-linto-2.2.0','link':'https://alphacephei.com/vosk/models/vosk-model-fr-0.6-linto-2.2.0.zip'}} # French
    ,'de':{'small':{'name':'vosk-model-small-de-0.15','link':'https://alphacephei.com/vosk/models/vosk-model-small-de-0.15.zip'},'medium':{'name':'vosk-model-de-0.21','link':'https://alphacephei.com/vosk/models/vosk-model-de-0.21.zip'},'large':{'name':'vosk-model-de-tuda-0.6-900k','link':'https://alphacephei.com/vosk/models/vosk-model-de-tuda-0.6-900k.zip'}} # German
    ,'es':{'small':{'name':'vosk-model-small-es-0.42','link':'https://alphacephei.com/vosk/models/vosk-model-small-es-0.42.zip'},'medium':{'name':'vosk-model-es-0.42','link':'https://alphacephei.com/vosk/models/vosk-model-es-0.42.zip'}} # Spanish
    ,'pt':{'small':{'name':'vosk-model-small-pt-0.3','link':'https://alphacephei.com/vosk/models/vosk-model-small-pt-0.3.zip'},'medium':{'name':'vosk-model-pt-fb-v0.1.1-20220516_2113','link':'https://alphacephei.com/vosk/models/vosk-model-pt-fb-v0.1.1-20220516_2113.zip'}} # Portuguese/Brazilian
    ,'gr':{'medium':{'name':'vosk-model-el-gr-0.7','link':'https://alphacephei.com/vosk/models/vosk-model-el-gr-0.7.zip'}} # Greek
    ,'tr':{'small':{'name':'vosk-model-small-tr-0.3','link':'https://alphacephei.com/vosk/models/vosk-model-small-tr-0.3.zip'}} # Turkish
    ,'vn':{'small':{'name':'vosk-model-small-vn-0.4','link':'https://alphacephei.com/vosk/models/vosk-model-small-vn-0.4.zip'},'medium':{'name':'vosk-model-vn-0.4','link':'https://alphacephei.com/vosk/models/vosk-model-vn-0.4.zip'}} # Vietnamese
    ,'it':{'small':{'name':'vosk-model-small-it-0.22','link':'https://alphacephei.com/vosk/models/vosk-model-small-it-0.22.zip'},'medium':{'name':'vosk-model-it-0.22','link':'https://alphacephei.com/vosk/models/vosk-model-it-0.22.zip'}} # Italian
    ,'nl':{'small':{'name':'vosk-model-small-nl-0.22','link':'https://alphacephei.com/vosk/models/vosk-model-small-nl-0.22.zip'},'medium':{'name':'vosk-model-nl-spraakherkenning-0.6','link':'https://alphacephei.com/vosk/models/vosk-model-nl-spraakherkenning-0.6.zip'}} # Dutch
    ,'ar':{'small':{'name':'vosk-model-ar-mgb2-0.4','link':'https://alphacephei.com/vosk/models/vosk-model-ar-mgb2-0.4.zip'},'medium':{'name':'vosk-model-ar-0.22-linto-1.1.0','link':'https://alphacephei.com/vosk/models/vosk-model-ar-0.22-linto-1.1.0.zip'}} # Arabic
    ,'fa':{'small':{'name':'vosk-model-small-fa-0.5','link':'https://alphacephei.com/vosk/models/vosk-model-small-fa-0.5.zip'},'medium':{'name':'vosk-model-fa-0.42','link':'https://alphacephei.com/vosk/models/vosk-model-fa-0.42.zip'}} # Farsi/Persian
    ,'ph':{'small':{'name':'vosk-model-tl-ph-generic-0.6','link':'https://alphacephei.com/vosk/models/vosk-model-tl-ph-generic-0.6.zip'}} # Filipino
    ,'uk':{'small':{'name':'vosk-model-small-uk-v3-small','link':'https://alphacephei.com/vosk/models/vosk-model-small-uk-v3-small.zip'},'medium':{'name':'vosk-model-uk-v3','link':'https://alphacephei.com/vosk/models/vosk-model-uk-v3.zip'}} # Ukrainian
    ,'sv':{'small':{'name':'vosk-model-small-sv-rhasspy-0.15','link':'https://alphacephei.com/vosk/models/vosk-model-small-sv-rhasspy-0.15.zip'}} # Swedish
    ,'ja':{'small':{'name':'vosk-model-small-ja-0.22','link':'https://alphacephei.com/vosk/models/vosk-model-small-ja-0.22.zip'},'medium':{'name':'vosk-model-ja-0.22','link':'https://alphacephei.com/vosk/models/vosk-model-ja-0.22.zip'}} # Japanese
    ,'hi':{'small':{'name':'vosk-model-small-hi-0.22','link':'https://alphacephei.com/vosk/models/vosk-model-small-hi-0.22.zip'},'medium':{'name':'vosk-model-hi-0.22','link':'https://alphacephei.com/vosk/models/vosk-model-hi-0.22.zip'}} # Hindi
    ,'pl':{'small':{'name':'vosk-model-small-pl-0.22','link':'https://alphacephei.com/vosk/models/vosk-model-small-pl-0.22.zip'}} # Polish
    ,'ko':{'small':{'name':'vosk-model-small-ko-0.22','link':'https://alphacephei.com/vosk/models/vosk-model-small-ko-0.22.zip'}} # Korean
    ,'tg':{'small':{'name':'vosk-model-small-tg-0.22','link':'https://alphacephei.com/vosk/models/vosk-model-small-tg-0.22.zip'},'medium':{'name':'vosk-model-tg-0.22','link':'https://alphacephei.com/vosk/models/vosk-model-tg-0.22.zip'}} # Tajik
    }
    def GetModel(language:str,size:str):
        if not perfectdict.has_key(VoskModels.supported_languages,language):
            assert VoskModels.supported_languages[language]
        model_=perfectdict.get_key(VoskModels.supported_languages[language],size)
        if model_ == None:
            return VoskModels.GetSmallerModel(language,size)
        return model_
    def AddVoskModel(name:str,language:str,link:str):
        if perfectdict.has_key(VoskModels.supported_languages,language):
            if perfectdict.has_key(VoskModels.supported_languages[language],'custom') and type(VoskModels.supported_languages[language]['custom']) == list:
                VoskModels.supported_languages[language]['custom'].append({'name':name,'link':link})
            else:
                VoskModels.supported_languages[language]['custom']=[]
                VoskModels.supported_languages[language]['custom'].append({'name':name,'link':link})
        else:
            VoskModels.supported_languages[language] = {}
            VoskModels.AddVoskModel(name,language,link)
        return True
    def GetSmallestModel(language:str):
        if not perfectdict.has_key(VoskModels.supported_languages,language):
            assert VoskModels.supported_languages[language]
        
        if perfectdict.has_key(VoskModels.supported_languages[language],'small'):
            return perfectdict.get_key(VoskModels.supported_languages[language],'small')
        elif perfectdict.has_key(VoskModels.supported_languages[language],'medium'):
            return perfectdict.get_key(VoskModels.supported_languages[language],'medium')
        elif perfectdict.has_key(VoskModels.supported_languages[language],'custom'):
            return perfectdict.get_key(VoskModels.supported_languages[language],'custom')[0]
        elif perfectdict.has_key(VoskModels.supported_languages[language],'large'):
            return perfectdict.get_key(VoskModels.supported_languages[language],'large')
        else:
            return None
    def GetBestModel(language:str):
        if not perfectdict.has_key(VoskModels.supported_languages,language):
            assert VoskModels.supported_languages[language]
        if perfectdict.has_key(VoskModels.supported_languages[language],'large'):
            return perfectdict.get_key(VoskModels.supported_languages[language],'large')
        elif perfectdict.has_key(VoskModels.supported_languages[language],'medium'):
            return perfectdict.get_key(VoskModels.supported_languages[language],'medium')
        elif perfectdict.has_key(VoskModels.supported_languages[language],'custom'):
            return perfectdict.get_key(VoskModels.supported_languages[language],'custom')[0]
        elif perfectdict.has_key(VoskModels.supported_languages[language],'small'):
            return perfectdict.get_key(VoskModels.supported_languages[language],'small')
        else:
            return None
    def GetCustomModels(language:str):
        if not perfectdict.has_key(VoskModels.supported_languages,language):
            assert VoskModels.supported_languages[language]
        if not perfectdict.has_key(VoskModels.supported_languages[language],'custom'):
            return None
        return perfectdict.get_key(VoskModels.supported_languages[language],'custom')
    def GetRandomModel(language:str):
        if not perfectdict.has_key(VoskModels.supported_languages,language):
            assert VoskModels.supported_languages[language]
        return VoskModels.supported_languages[language][_random.choice(list(VoskModels.supported_languages[language].keys()))]
    def GetMediumModel(language:str):
        if not perfectdict.has_key(VoskModels.supported_languages,language):
            assert VoskModels.supported_languages[language]
        return perfectdict.get_key(VoskModels.supported_languages[language],'medium')
    def GetSmallerModel(language:str,weight:str):
        if not perfectdict.has_key(VoskModels.supported_languages,language):
            assert VoskModels.supported_languages[language]
        if weight == 'large':
            return perfectdict.get_key(VoskModels.supported_languages[language],'medium')
        elif weight == 'medium':
            return perfectdict.get_key(VoskModels.supported_languages[language],'small')
        elif weight == 'small':
            return perfectdict.get_key(VoskModels.supported_languages[language],'small')
        else:
            assert ValueError(f"Unknown weight {weight}, unable to get smaller model.")
    def GetLargerModel(language:str,weight:str):
        if not perfectdict.has_key(VoskModels.supported_languages,language):
            assert VoskModels.supported_languages[language]
        if weight == 'small':
            return perfectdict.get_key(VoskModels.supported_languages[language],'medium')
        elif weight == 'medium':
            return perfectdict.get_key(VoskModels.supported_languages[language],'large')
        elif weight == 'large':
            return perfectdict.get_key(VoskModels.supported_languages[language],'large')
        else:
            assert ValueError(f"Unknown weight {weight}, unable to get larger model.")
    def GetAvailableSizes(language:str):
        if not perfectdict.has_key(VoskModels.supported_languages,language):
            assert VoskModels.supported_languages[language]
        return perfectdict.get_keys(VoskModels.supported_languages[language])
    def GetAvailableLangs():
        return perfectdict.get_keys(VoskModels.supported_languages)

class _VMM_CUSTOM_DOWNLOAD_MANAGER:
    def __init__(self,download_obj:_download,model:dict):
        self.obj=download_obj
        self.model = model
    def status(self): #paused, downloading, finished
        return self.obj.get_status()
    def start(self):
        self.obj.start(False)
    def restart(self,auto_start=False):
        self.obj.retry()
        self.obj.start()
        return True
    def is_finished(self):
        if self.obj.isFinished():
            return True
        return False
    def finished_and_successful(self):
        return self.obj.isFinished() and self.obj.isSuccessful()
    def string_full_status(self):
        try:
            return f"Downloading ({self.obj.get_speed(True)}) {self.obj.get_progress_bar(30)} {self.obj.get_dl_size()}/{self.obj.get_final_filesize()} {self.obj.get_eta()}"
        except Exception as err:
            print(f"ERROR: {err}")
            return ""
    def after_download(self):
        def _get_only_folder(folder_path:str):
            folder = Path(folder_path)

            # Ensure the path is a directory
            if not folder.exists() or not folder.is_dir():
                raise FileNotFoundError(f"The path '{folder_path}' does not exist or is not a directory.")
            
            # List all subdirectories in the given folder
            subfolders = [f for f in folder.iterdir() if f.is_dir()]
            
            # Check if exactly one folder exists
            if len(subfolders) == 0:
                raise ValueError(f"No folders found in '{folder_path}'.")
            elif len(subfolders) > 1:
                raise ValueError(f"More than one folder found in '{folder_path}': {', '.join(str(f) for f in subfolders)}")
            
            # Return the single folder
            return subfolders[0]
        
        downloaded = self.obj
        prints=False
        model=self.model
        if downloaded.isFinished():
            if prints:print("Model download completed. installing model...")
            zfile=downloaded.get_dest()
            _ZipFile(zfile).extractall(str(Path(VMM.models_path).joinpath(model['name'])))
            Path(VMM.models_path).joinpath('temp').mkdir(exist_ok=True)
            try:
                orgn=_get_only_folder(Path(VMM.models_path).joinpath(model['name'])).name
                _shutil.move(_get_only_folder(Path(VMM.models_path).joinpath(model['name'])),Path(VMM.models_path).joinpath("temp"))
                temp_folder_name= _get_only_folder(Path(VMM.models_path).joinpath('temp')).name
                _shutil.rmtree(Path(VMM.models_path).joinpath(model['name']))
                _shutil.move(_get_only_folder(Path(VMM.models_path).joinpath('temp')),Path(VMM.models_path))
                _shutil.rmtree(Path(VMM.models_path).joinpath('temp'))
                Path(VMM.models_path).joinpath(temp_folder_name).rename(Path(VMM.models_path).joinpath(model['name']))
            except Exception as err: raise SystemError(err)
            if Path(VMM.models_path).joinpath(orgn):
                if prints:print("Model successfully installed.")
                return True
            else:
                raise SystemError("System failed to install the model.")
        else:
            if prints:print("Model download failed. quiting...")
            return False
    

class VMM: # Vosk Model Manager
    models_path = str(Path().home().joinpath(".omgl/").joinpath("vosk_models/"))
    def change_models_path(newpath:str):
        VMM.models_path = str(Path(newpath).absolute())
        return True
    def get_model_path(model:dict):
        return str(Path(VMM.models_path).joinpath(model['name']))
    def smart_download(model:dict,prints:bool=False):
        def _get_only_folder(folder_path:str):
            folder = Path(folder_path)
    
            # Ensure the path is a directory
            if not folder.exists() or not folder.is_dir():
                raise FileNotFoundError(f"The path '{folder_path}' does not exist or is not a directory.")
            
            # List all subdirectories in the given folder
            subfolders = [f for f in folder.iterdir() if f.is_dir()]
            
            # Check if exactly one folder exists
            if len(subfolders) == 0:
                raise ValueError(f"No folders found in '{folder_path}'.")
            elif len(subfolders) > 1:
                raise ValueError(f"More than one folder found in '{folder_path}': {', '.join(str(f) for f in subfolders)}")
            
            # Return the single folder
            return subfolders[0]
        
        if prints:
            print("Checking to see if model is already installed...")
        if Path(VMM.models_path).joinpath(model['name']+'.zip').exists():
            zfile=str(Path(VMM.models_path).joinpath(model['name']+'.zip'))
            _ZipFile(zfile).extractall(str(Path(VMM.models_path).joinpath(model['name'])))
            Path(VMM.models_path).joinpath('temp').mkdir(exist_ok=True)
            try:
                orgf=_get_only_folder(Path(VMM.models_path).joinpath(model['name']))
                orgn=_get_only_folder(Path(VMM.models_path).joinpath(model['name'])).name
                _shutil.move(_get_only_folder(Path(VMM.models_path).joinpath(model['name'])),Path(VMM.models_path).joinpath("temp"))
                _shutil.rmtree(Path(VMM.models_path).joinpath(model['name']))
                temp_folder_name= _get_only_folder(Path(VMM.models_path).joinpath('temp')).name
                _shutil.move(_get_only_folder(Path(VMM.models_path).joinpath('temp')),Path(VMM.models_path))
                _shutil.rmtree(Path(VMM.models_path).joinpath('temp'))
                Path(VMM.models_path).joinpath(temp_folder_name).rename(Path(VMM.models_path).joinpath(model['name']))
            except Exception as err: raise SystemError(err)
            if Path(VMM.models_path).joinpath(orgn):
                if prints:print("Model successfully installed.")
                return True
            else:
                raise SystemError("System failed to install the model.")
        if Path(VMM.models_path).joinpath(model['name']).exists():
            return True
        if prints:print("model is not installed, downloading model...")
        dest=str(Path(VMM.models_path).joinpath(model['name']+'.zip'))

        downloaded=_download(model['link'],dest,False)
        return _VMM_CUSTOM_DOWNLOAD_MANAGER(downloaded,model)
        
    def install_model(model:dict,prints=False):
        def _get_only_folder(folder_path:str):
            folder = Path(folder_path)
    
            # Ensure the path is a directory
            if not folder.exists() or not folder.is_dir():
                raise FileNotFoundError(f"The path '{folder_path}' does not exist or is not a directory.")
            
            # List all subdirectories in the given folder
            subfolders = [f for f in folder.iterdir() if f.is_dir()]
            
            # Check if exactly one folder exists
            if len(subfolders) == 0:
                raise ValueError(f"No folders found in '{folder_path}'.")
            elif len(subfolders) > 1:
                raise ValueError(f"More than one folder found in '{folder_path}': {', '.join(str(f) for f in subfolders)}")
            
            # Return the single folder
            return subfolders[0]
        
        if prints:
            print("Checking to see if model is already installed...")
        if Path(VMM.models_path).joinpath(model['name']+'.zip').exists():
            zfile=str(Path(VMM.models_path).joinpath(model['name']+'.zip'))
            _ZipFile(zfile).extractall(str(Path(VMM.models_path).joinpath(model['name'])))
            Path(VMM.models_path).joinpath('temp').mkdir(exist_ok=True)
            try:
                orgf=_get_only_folder(Path(VMM.models_path).joinpath(model['name']))
                orgn=_get_only_folder(Path(VMM.models_path).joinpath(model['name'])).name
                _shutil.move(_get_only_folder(Path(VMM.models_path).joinpath(model['name'])),Path(VMM.models_path).joinpath("temp"))
                _shutil.rmtree(Path(VMM.models_path).joinpath(model['name']))
                temp_folder_name= _get_only_folder(Path(VMM.models_path).joinpath('temp')).name
                _shutil.move(_get_only_folder(Path(VMM.models_path).joinpath('temp')),Path(VMM.models_path))
                _shutil.rmtree(Path(VMM.models_path).joinpath('temp'))
                Path(VMM.models_path).joinpath(temp_folder_name).rename(Path(VMM.models_path).joinpath(model['name']))
            except Exception as err: raise SystemError(err)
            if Path(VMM.models_path).joinpath(orgn):
                if prints:print("Model successfully installed.")
                return True
            else:
                raise SystemError("System failed to install the model.")
        if Path(VMM.models_path).joinpath(model['name']).exists():
            return True
        if prints:print("model is not installed, downloading model...")
        dest=str(Path(VMM.models_path).joinpath(model['name']+'.zip'))

        downloaded=_download(model['link'],dest)
        downloaded.start()
        downloaded.wait()
        if downloaded.isFinished():
            if prints:print("Model download completed. installing model...")
            zfile=downloaded.get_dest()
            _ZipFile(zfile).extractall(str(Path(VMM.models_path).joinpath(model['name'])))
            Path(VMM.models_path).joinpath('temp').mkdir(exist_ok=True)
            try:
                orgn=_get_only_folder(Path(VMM.models_path).joinpath(model['name'])).name
                _shutil.move(_get_only_folder(Path(VMM.models_path).joinpath(model['name'])),Path(VMM.models_path).joinpath("temp"))
                temp_folder_name= _get_only_folder(Path(VMM.models_path).joinpath('temp')).name
                _shutil.rmtree(Path(VMM.models_path).joinpath(model['name']))
                _shutil.move(_get_only_folder(Path(VMM.models_path).joinpath('temp')),Path(VMM.models_path))
                _shutil.rmtree(Path(VMM.models_path).joinpath('temp'))
                Path(VMM.models_path).joinpath(temp_folder_name).rename(Path(VMM.models_path).joinpath(model['name']))
            except Exception as err: raise SystemError(err)
            if Path(VMM.models_path).joinpath(orgn):
                if prints:print("Model successfully installed.")
                return True
            else:
                raise SystemError("System failed to install the model.")
        else:
            if prints:print("Model download failed. quiting...")
            return False
    def clear_cache(are_you_sure:bool=False): # This function will remove every zip archive file of models_path folder be ware
        if not are_you_sure:return False
        zip_files = list(Path(VMM.models_path).rglob('*.zip'))
        for zip_file in zip_files:
            if 'vosk' in str(zip_file.absolute()):
                try:
                    zip_file.unlink()
                except:...
        return True
    def clear_folders(are_you_sure:bool=False):
        import warnings,time
        if are_you_sure: warnings.warn(f"You are about to delete all models in the {VMM.models_path}. Please cancel with CTRL+C")
        else:return False
        time.sleep(5)
        print("WARNING: DELETING ALL MODELS, PLEASE DO NOT CANCEL.")
        for item in Path(VMM.models_path).iterdir():
            if item.is_dir():
                _shutil.rmtree(item)
                print(f"Deleted model folder: {item}")
    def _completely_clear_cache(are_you_sure:bool=False):
        import warnings,time
        if are_you_sure: warnings.warn(f"You are about to delete all files/models in the {VMM.models_path}. Please cancel with CTRL+C")
        else:return False
        time.sleep(5)
        print("WARNING: DELETING ALL FILES, PLEASE DO NOT CANCEL.")
        for item in Path(VMM.models_path).iterdir():
            if item.is_dir():
                _shutil.rmtree(item)
                print(f"Deleted folder: {item}")
            else:
                print(f"Deleted file: {item}")
                item.unlink()
        
    def ez_install_model(model_lang:str=Languages.English,model_size:str=VoskModelSize.Small):
        model=VoskModels.GetModel(model_lang,model_size)
        VMM.install_model(model)
    def model_exists(model:dict):
        if Path(VMM.models_path).joinpath(model['name']).exists():
            return True
        return False
    def ez_model_exists(model_lang:str,model_size:str=VoskModelSize.Small):
        return True if perfectdict.has_key(VoskModels.supported_languages,model_lang) and perfectdict.has_key(VoskModels.supported_languages[model_lang],model_size) else False


class WhisperModels:
    available_models=_wa()
    def _winit():
        for model in WhisperModels.available_models:
            setattr(WhisperModels,model.capitalize().replace('-','').replace('.','_'),model)
WhisperModels._winit()


import tools
import asyncio


if __name__ == "__main__":
    assert tools.run_py("print('This is working just right')") == {'run_status':'Successful',"error_message":"<No Error>"}
    assert tools.run_py("gimmesomeerror_SADS!()*()")['run_status'] == "Failed"
    assert list(tools.get_weather("New York")) == ['location', 'temperature', 'unit']
    


from ..tools.dicttools import perfectdict

class SelectKeyExists(Exception):...
class SelectKeyDoesntExists(Exception):...
class InvalidSelectionValue(Exception):...


class Selector:
    def __init__(self,selector_name:str="SELECTOR"):
        self._SELECTOR_ = selector_name
    def __repr__(self):
        return f"<{self._SELECTOR_} {str(id(self))[-6:]}>"
    def _first_run(self,initial_config):
        if not hasattr(self,'_select_options'):
            self._select_options = initial_config
        return

class SelectKey:
    def __init__(self,name:str=None):
        if name == None:
            self.name=id(self)
        else:
            self.name = name
    def __call__(self):
        return self.name
    def __repr__(self):
        return f"<SelectKey name={self.name}>"

class DictSelector(Selector):
    _initial_config = {}
    def __init__(self):
        super().__init__("DictSelector")
        self._first_run(DictSelector._initial_config)
        self._recognize_by = 'name'
    def load_from_dict(self,dict_:dict):
        self._select_options = dict_
        return self
    def _safe_value(self,value):
        if type(value) in [int,str,bool,dict,list,float,complex,range,tuple,set,frozenset,bytes,bytearray,memoryview,type(None)]:
            return value
        else:
            raise InvalidSelectionValue(f"SAFEVALUE:-> Value '{value}' is not a valid/supported format for DictSelector, use 'bypass_safe_value' property to disable it.")
    def _safe_check(self,selkey:SelectKey):
        if perfectdict.has_key(self._select_options,selkey):
            raise SelectKeyExists(f"SAFECHECK:-> Key '{selkey}' exists in the DictSelector.")
        return True
    def add_selection(self,selection_key:SelectKey,selection_value:any,bypass_safe_check:bool=False,bypass_safe_value:bool=False):
        self._safe_check(selection_key) if bypass_safe_check == False else None
        self._select_options[selection_key] = self._safe_value(selection_value) if bypass_safe_value == False else selection_value
        self._select_options[selection_key] = selection_value
        return self
    def delete_selection(self,selection_key:SelectKey):
        if isinstance(selection_key,tuple) or isinstance(selection_key,list):
            for key in selection_key:
                self.delete_selection(key)
            return self
        if perfectdict.has_key(self._select_options,selection_key):
            del self._select_options[selection_key]
        return self
    @property
    def all(self):
        return self.select_all()
    def select_all(self):
        return self._select_options
    def select(self,selection_key:SelectKey):
        if selection_key == "*" and perfectdict.nhas_key(self._select_options,selection_key):
            return self._select_options
        if perfectdict.nhas_key(self._select_options,selection_key) and not perfectdict.wkey_has_prop(self._select_options,self._recognize_by,selection_key):
            raise SelectKeyDoesntExists(f"Key '{selection_key}' doesn't exist in the DictSelector.")
        if perfectdict.has_key(self._select_options,selection_key):
            return self._select_options[selection_key]
        if perfectdict.wkey_has_prop(self._select_options,self._recognize_by,selection_key):
            return self._select_options[perfectdict.gkey_has_prop(self._select_options,self._recognize_by,selection_key)]
        return self._select_options[selection_key]
    def __call__(self,selection_key:SelectKey=None,selection_value:any=None):
        if selection_key == None and selection_value == None:
            return self
        elif selection_key != None and selection_value == None:
            return self.select(selection_key)
        else: 
            return self.add_selection(selection_key,selection_value)
    def __matmul__(self,selection_key:SelectKey):
        return self.delete_selection(selection_key)
    def __iter__(self):
        for item in range(len(perfectdict.get_keys(self._select_options))):
            yield self._select_options[item]
    def __gt__(self,item:tuple):
        if type(item) is not tuple:return self.__call__(item)
        return self.__call__(*item)
    def __lt__(self,item:tuple):
        if type(item) is not tuple:return self.__call__(item)
        return self.__call__(*item)
    def __getitem__(self,item:SelectKey):
        return self.__call__(item)


from openai import OpenAI
import json


class BasicAPI:
    def __init__(self,base_url:str,apikey:str,model_name:str,tools_use:bool=False):
        self.client = OpenAI(base_url=base_url,api_key=apikey)
        self.model = model_name
        self.max_tokens = 8000
        self.messages=[]
        self.tools = []
        self.tools_use = tools_use
    def system_message(self,content):
        self.messages.append({"role":"system","content":content})
        return True
    def user_message(self,content):
        self.messages.append({"role":"user","content":content})
        return True
    def set_tools(self,tools_list:list,tool_handle):
        self.tools_use = True
        self.tools = tools_list
        self.tool_handle_function = tool_handle
        return True
    def assistant_message(self):
        if not self.tools_use:
            comp = self.client.chat.completions.create(messages=self.messages,model=self.model,max_tokens=self.max_tokens)
            response = comp.choices[0].message
            self.messages.append({"role":response.role,"content":response.content})
            return response.content
        else:
            comp = self.client.chat.completions.create(messages=self.messages,model=self.model,max_tokens=self.max_tokens,tools=self.tools,tool_choice="auto")
            response = comp.choices[0].message
            if response.tool_calls:
                tool_call_id = response.tool_calls[0].id
                func_response=self.tool_handle_function(response.tool_calls.function.name,json.loads(response.tool_calls.function.arguments))
                self.messages.append(dict(response))
                self.messages.append({
                    'role':'tool',
                    'content':json.dumps(func_response),
                    'tool_call_id':tool_call_id
                })
                return self.assistant_message()
            else:
                self.messages.append({"role":response.role,"content":response.content})
                return response.content


class BasicCustom_API:
    def __init__(self,custom_api_client:OpenAI,model_name:str,tools_use:bool=False):
        self.client = custom_api_client
        self.model = model_name
        self.max_tokens = 8000
        self.messages=[]
        self.tools = []
        self.tools_use = tools_use
    def system_message(self,content):
        self.messages.append({"role":"system","content":content})
        return True
    def user_message(self,content):
        self.messages.append({"role":"user","content":content})
        return True
    def set_tools(self,tools_list:list,tool_handle):
        self.tools_use = True
        self.tools = tools_list
        self.tool_handle_function = tool_handle
        return True
    def assistant_message(self):
        if not self.tools_use:
            comp = self.client.chat.completions.create(messages=self.messages,model=self.model,max_tokens=self.max_tokens)
            response = comp.choices[0].message
            self.messages.append({"role":response.role,"content":response.content})
            return response.content
        else:
            comp = self.client.chat.completions.create(messages=self.messages,model=self.model,max_tokens=self.max_tokens,tools=self.tools,tool_choice="auto")
            response = comp.choices[0].message
            if response.tool_calls:
                tool_call_id = response.tool_calls[0].id
                func_response=self.tool_handle_function(response.tool_calls.function.name,json.loads(response.tool_calls.function.arguments))
                self.messages.append(dict(response))
                self.messages.append({
                    'role':'tool',
                    'content':json.dumps(func_response),
                    'tool_call_id':tool_call_id
                })
                return self.assistant_message()
            else:
                self.messages.append({"role":response.role,"content":response.content})
                return response.content



from openai import OpenAI
import speech_recognition as _SR


class oWhisper:
    def __init__(self,apikey:str,language:str='en',model:str='whisper-1'):
        self._openai = OpenAI(api_key=apikey)
        self.model=model
        self.language=language
    def transcribe(self,raw_audio:str):
        return self._openai.audio.transcriptions.create(file=raw_audio,model=self.model,language=self.language)['text']


class GoogleCloudAPI:
    def __init__(self,json_creds:str,language:str='en'):
        self.json_creds=json_creds
        self.language=language
        self._R = _SR.Recognizer()
    def recognize(self,audio_data):
        return self._R.recognize_google_cloud(audio_data,credentials_json=self.json_creds,language=self.language)




from ..listtools.listing import aget_only_item
class _wTempClass:name=None
def has_key(dict_:dict,key:any):return True if key in dict_.keys() else False
def nhas_key(dict_:dict,key:any):return False if key in dict_.keys() else True
def has_value(dict_:dict,value:any):return True if value in dict_.values() else False
def nhas_value(dict_:dict,value:any):return False if value in dict_.values() else True
def get_key(dict_:dict,key:any):return dict_[key] if has_key(dict_,key) else None
def get_keys(dict_:dict):return list(dict_.keys())
def get_values(dict_:dict):return list(dict_.values())
def has_keys(dict_:dict,keys:list):return True if all([key in keys for key in dict_.keys()]) else False
def nhas_keys(dict_:dict,keys:list):return False if all([key in keys for key in dict_.keys()]) else True
def nhas_atleast_onekey(dict_:dict,keys:list):return False if any([key in keys for key in dict_.keys()]) else True
def has_atleast_onekey(dict_:dict,keys:list):return True if any([key in keys for key in dict_.keys()]) else False
def wkey_has_prop(dict_:dict,property_name:str,property_value)->bool:ks,chks=[k for k in list(dict_.keys())],[hasattr(k,property_name) for k in list(dict_.keys())];return any([True if getattr(ks[ind] if isinstance(ind,int) else _wTempClass,'name') == property_value else False for ind in [i if chks[i] == True else None for i in range(len(chks))]])
def gkey_has_prop(dict_:dict,property_name:str,property_value)->bool:ks,chks=[k for k in list(dict_.keys())],[hasattr(k,property_name) for k in list(dict_.keys())];return aget_only_item([ks[ind] if getattr(ks[ind] if isinstance(ind,int) else _wTempClass,'name') == property_value else False for ind in [i if chks[i] == True else None for i in range(len(chks))]])
def key_ret(dict_:dict,key:any):return get_key(dict_,key) if has_key(dict_,key) else None


class workspace:...
class ListTools():
    def __init__(self,list):
        self._list = list
    def ToString(self)->str:
        out = ''
        for i in self._list:
            out += str(i)
        return out
    def if_list_in_list(self):
        for i in self._list:
            if type(i) == list:
                return True
        return False
    def if_type_in_list(self,type_:type):
        for i in self._list:
            if type(i) == type_:
                return True
        return False
    def GAItemlist(self):
        GAItem(self._list)
        return GAItem_out()


def get_only_item(ls:list):
    for i in ls:
        if i != None:
            return i
    return None
def aget_only_item(ls:list):
    for i in ls:
        if i != None and i != False:
            return i
    return None

def ToString(ls:list)->str:
    out = ''
    for i in ls:
        out += str(i)
    return out
 

def if_list_in_list(ls:list):
    for i in ls:
        if type(i) == list:
            return True
    return False

def if_type_in_list(type_:type,ls:list):
    for i in ls:
        if type(i) == type_:
            return True
    return False



def GAItem(ls:list)->list:
    def if_type(ls_):
        for i in ls_:
            if type(i) is list:
                return False
        return True
                
    out = []
    check_again = False
    for i in ls:
        if type(i) != list:
            out.append(i)
        else:
            check_again = True
            for x in i:
                out.append(x)
    for i in out:
        if type(i) != list:
            check_again = False
        else:
            check_again = True
            
    if check_again:
        GAItem(out)
    else:
        
        if not if_type(out):
            GAItem(out)
        workspace.GA_out = out
        return out

def GAItem_out():
    return workspace.GA_out
    
def GAItemlist(ls):
    GAItem(ls)
    return GAItem_out()

def alist():
    return []
def srange(start,end,step=1):
    if start == abs(start):
        for i in range(start,end,step):
            yield i
    if type(start)!=int or type(end)!=int:
        raise TypeError('Invalid Types in use.')
    if step > end:
        yield start
    else:
        eq = start
        yield eq
        while True:
            if eq == 0:
                for i in range(eq+1,end,step):
                    if eq >= end:
                        break
                    yield i
                break
            if eq >= end:
                break
            eq+=step
            yield eq
def tlist_for_each(ls,enum = lambda x:x ):
    out = []
    
    for item in ls:
        out.append(enum(item))
    return ListTools(out)

def list_for_inner(gen):
    out = []
    for i in gen:
        out.append(i)
    return out

def list_for_each(ls,enum = lambda x:x ):
    out = []
    for item in ls:
        out.append(enum(item))
    yield out



import zipfile
from pathlib import Path

def extract_folder_from_zip(zip_file_path: str, folder_in_zip: str, output_folder: str):
    """
    Extracts all files from a specific folder inside a zip archive to another folder.
    
    Parameters:
        zip_file_path (str): Path to the zip file.
        folder_in_zip (str): The folder inside the zip file to extract.
        output_folder (str): The destination folder for extracted files.
    """
    zip_path = Path(zip_file_path)
    output_path = Path(output_folder)
    
    # Ensure the zip file exists
    if not zip_path.exists():
        raise FileNotFoundError(f"Zip file '{zip_file_path}' not found.")
    
    # Ensure the output folder exists
    output_path.mkdir(parents=True, exist_ok=True)
    
    # Open the zip file
    with zipfile.ZipFile(zip_path, 'r') as zip_file:
        # Iterate through the files in the zip file
        for file in zip_file.namelist():
            # Check if the file is within the specified folder
            if file.startswith(f"{folder_in_zip}/") and not file.endswith('/'):  # Skip directories
                # Extract file to the output folder
                destination = output_path / Path(file).name
                with zip_file.open(file) as source, open(destination, "wb") as target:
                    target.write(source.read())
                print(f"Extracted: {file} to {destination}")


import requests
import json
import pyaudio
from pydub import AudioSegment
from io import BytesIO


def play_mp3_bytes(mp3_data: bytes):
    """
    Play MP3 audio data from a bytes object using PyAudio.
    
    :param mp3_data: MP3 audio data as bytes
    """
    # Decode MP3 bytes to raw audio data
    audio = AudioSegment.from_file(BytesIO(mp3_data), format="mp3")

    # Extract raw audio data and parameters
    raw_data = audio.raw_data
    sample_width = audio.sample_width
    frame_rate = audio.frame_rate
    channels = audio.channels
    
    # Initialize PyAudio
    p = pyaudio.PyAudio()
    
    # Open a PyAudio stream
    stream = p.open(format=p.get_format_from_width(sample_width),
                    channels=channels,
                    rate=frame_rate,
                    output=True)
    
    # Play the raw audio data
    stream.write(raw_data)
    
    # Stop and close the stream
    stream.stop_stream()
    stream.close()
    
    # Terminate PyAudio
    p.terminate()

class MurfAI:
    def __init__(self,apikey:str):
        self.apikey=apikey
        self.url = "https://api.murf.ai/v1/speech/generate"
    def get_file(self,text:str,lang:str="en-US-natalie",rate:int=0,pitch:int=0,samplerate:int=48000,style:str="Conversational"):
        payload=json.dumps({
  "voiceId": lang,
  "style": style,
  "text": text,
  "rate": rate,
  "pitch": pitch,
  "sampleRate": samplerate,
  "format": "MP3",
  "channelType": "MONO",
  "pronunciationDictionary": {},
  "encodeAsBase64": False,
  "variation": 1,
  "audioDuration": 0,
  "modelVersion": "GEN2"
})
        headers = {
  'Content-Type': 'application/json',
  'Accept': 'application/json',
  'api-key': self.apikey
}
        response=requests.request("POST",self.url,headers=headers,data=payload)
        return response.text
    def speak(self,text:str,lang:str="en-US-natalie",rate:int=0,pitch:int=0,samplerate:int=48000,style:str="Promo"):
        self.audioFile=json.loads(self.get_file(text,lang,rate,pitch,samplerate,style))['audioFile']
        audioData = requests.get(self.audioFile).content
        self.last_audio = audioData
        return play_mp3_bytes(audioData)

class Playht:
    def __init__(self,userid:str,secretkey:str):
        self._userid = userid
        self._secretkey = secretkey
        self.voiceid="s3://voice-cloning-zero-shot/2a7ddfc5-d16a-423a-9441-5b13290998b8/novasaad/manifest.json"
    def speak(self,text:str):
        headers={
            "X-USER-ID":self._userid,
            "AUTHORIZATION":self._secretkey,
            "accept":"audio/mpeg",
            "content-type":"application/json"
        }
        payload={
            "text":text,
            "voice_engine":"Play3.0",
            "voice":self.voiceid,
            "output_format":"mp3"
        }

        self.audioData=requests.request("POST","https://api.play.ht/api/v2/tts/stream",headers=headers,data=json.dumps(payload)).content
        return play_mp3_bytes(self.audioData)

    def get_voices(self):
        headers={
            "X-USER-ID":self._userid,
            "AUTHORIZATION":self._secretkey,
            "accept":"application/json"
        }
        return requests.get("https://api.play.ht/api/v2/voices",headers=headers).json()




import json
import json.tool
import json.scanner
import requests
import python_weather
import asyncio
import serpapi
from pathlib import Path
import subprocess
from .tools_base import *
from ..tools.dicttools import perfectdict

# Built in functions ---------------------------------------------
def convert_path(path:str): # Converts linux path to windows path
    return str(Path(path).absolute())
def init_inuse_tools_import():
    global m
    import win32ui
    import windowsapps
    import winapps
    import webbrowser
    import subprocess
    import cython
    m.win32ui = win32ui
    m.windowsapps = windowsapps
    m.winapps = winapps
    m.webbrowser=webbrowser
    m.subprocess=subprocess
    m.cython = cython
#-----------------------------------------------------------------


global ALL_AVAILABLE_TOOLS
_debug = True

FunctionConfigs = {"search_web":{'apikey':None}}

class m:...

class XConfig:
    def set_search_web_apikey(apikey:str):
        FunctionConfigs['search_web']['apikey'] = apikey
        return True
    def add_search_web():
        ALL_AVAILABLE_TOOLS.append(gen_func("search_web",required=["query"],query=xtype(str,"the query for web search example: 'Top premiere leagues'"),search_engine=xtype(str,"The search engine to use for search example: 'google'",enum=['google','bing']),country_code=xtype(str,"The country code for regional search default:'US'"))) # search_web(query:str,search_engine:str='google',country_code:str="US")
        return True
    def add_tool(generated_func:dict):
        ALL_AVAILABLE_TOOLS.append(generated_func)

def OpenApp(appname:str)->bool:
    try:
        m.windowsapps.open_app(appname) # Uses a library to search and open application
        return {"OpenApplicationStatus":"Successful","RequestedApplicationName":appname} # returns true if the application is found and launched 
    except:
        return {"OpenApplicationStatus":"Failed/Probaby not found","RequestedApplicationName":appname} # returns false if the application is not found

def ListInstalledAppsName()->list:
    output=[] # the return variable
    for app in list(m.winapps.list_installed()): # loops through installed applications using "winapps" library
       output.append(app.name) # appends only the name of the applications installed
    return {"ApplicationsList":output} # returns the names of the applications installed

def UninstallApp(appname:str)->bool:
    if m.winapps.search_installed(appname)==[]:return False # Returns false if installed application is not found
    
    ask=m.win32ui.MessageBox(f"Do you really want to uninstall {appname}?","AI Assistant",4) # makes sure if user wants their application to be deleted
    if ask == 6:
        try:
            m.winapps.uninstall(appname) # uninstalls the application
            return {"UninstallAppRequest":"Successful","RequestedApplication":appname}
        except:
            return {"UninstallAppRequest":"Failed","RequestedApplication":appname,"Reason":"Probably due to being unable to locate the application or application doesn't exist."}
    else:
        return {"UninstallAppRequest":"Failed","RequestedApplication":appname,"Reason":"User rejected the verification to uninstall their application"}

def MessageBox(message:str,title:str,style:int=0)->dict:
    _= m.win32ui.MessageBox(message,title,style) # Shows MessageBox with arguements
    # checks for codes for answers and returns simpler answers
    if _ == 1:
        return {"MessageBoxStatus":"Shown_Successfully","MessageBoxReturnCode":_,"MessageBoxReturnCode_Translation":"OK_BUTTON"}
    elif _ == 2:
        return {"MessageBoxStatus":"Shown_Successfully","MessageBoxReturnCode":_,"MessageBoxReturnCode_Translation":"CANCEL_BUTTON"}
    elif _ == 3:
        return {"MessageBoxStatus":"Shown_Successfully","MessageBoxReturnCode":_,"MessageBoxReturnCode_Translation":"ABORT_BUTTON"}
    elif _ == 4:
        return {"MessageBoxStatus":"Shown_Successfully","MessageBoxReturnCode":_,"MessageBoxReturnCode_Translation":"RETRY_BUTTON"}
    elif _ == 5:
        return {"MessageBoxStatus":"Shown_Successfully","MessageBoxReturnCode":_,"MessageBoxReturnCode_Translation":"IGNORE_BUTTON"}
    elif _ == 6:
        return {"MessageBoxStatus":"Shown_Successfully","MessageBoxReturnCode":_,"MessageBoxReturnCode_Translation":"YES_BUTTON"}
    elif _ == 7:
        return {"MessageBoxStatus":"Shown_Successfully","MessageBoxReturnCode":_,"MessageBoxReturnCode_Translation":"NO_BUTTON"}
    elif _ == 10:
        return {"MessageBoxStatus":"Shown_Successfully","MessageBoxReturnCode":_,"MessageBoxReturnCode_Translation":"TRYAGAIN_BUTTON"}
    elif _ == 11:
        return {"MessageBoxStatus":"Shown_Successfully","MessageBoxReturnCode":_,"MessageBoxReturnCode_Translation":"CONTINUE_BUTTON"}
    else:
        return {"MessageBoxStatus":"Suspended","MessageBoxReturnCode":_,"MessageBoxReturnCode_Translation":"UNKNOWN"}

def OpenUrl(url:str): 
    m.webbrowser.open(url) # simply opens a url in browser
    return {
        "OpenUrlStatus":"Successful",
        "UrlOpened":url
    }
def OpenPath(path:str):
    # simply opens a path in windows explorer
    try:
        _=m.subprocess.Popen(["explorer.exe",convert_path(path)])
        return {
        "OpenPathStatus":"Successful",
        "Args":_.args
    }
    except Exception as e:
        return {
            "OpenPathStatus":"Failed",
            "PyError_Message":str(e)
        }

def cython_inline(inline_text:str):
    try:
        out=m.cython.inline(inline_text)
        return {"cython_inline_RunStatus":"Successful","CythonInlineOutput":str(out)}
    except Exception as e: 
        return {"cython_inline_RunStatus":"Failed","CythonInlineErrorMessage":str(e)}

# ______________________________________________________________________________________________________________
def run_py(python_code:str):
    run_status = "Successful"
    error_message = "<No Error>"
    try:
        exec(python_code)
    except Exception as e:
        run_status = "Failed"
        error_message=e
    return {"run_status":run_status,"error_message":error_message}

def run_cmd(cmd_code:str):
    _=subprocess.getstatusoutput(cmd_code)
    run_status = "Successful" if _[0] == 0 else "Failed" 
    return {"run_status":run_status,"output":_[1]}

async def _get_weather(location:str,unit:str='celcius'):
    cli=python_weather.Client(unit=python_weather.constants.METRIC if unit == "celcius" else python_weather.constants.IMPERIAL)
    weather=await cli.get(location=location)
    return {"location":weather.location,"temperature":weather.temperature,"unit":unit,"datetime":weather.datetime.strftime("%Y-%m-%d %H:%M:%S"),"additional_data":{"kind":str(weather.kind),"description":weather.description,"country":weather.country,"coordinates":list(weather.coordinates)}}

def get_weather(location:str,unit:str='celcius'):
    if asyncio.get_event_loop().is_closed():
        asyncio.set_event_loop(asyncio.new_event_loop())
    return asyncio.get_event_loop().run_until_complete(_get_weather(location=location,unit=unit))

def search_web(query:str,search_engine:str='google',country_code:str="US"): # Searches the web with English language results
    if FunctionConfigs['search_web']['apikey'] == None:
        return {"Error":"Failed","Further_Details_DEV":"User hasn't set any api key for search_web option, reject them or guide them."}
    if search_engine.lower() == 'google':
        results = {"query":query,"search_engine":search_engine,"search_results":{},"discussions_and_forums":{},"related_questions":{}}
        try:
            json_results=serpapi.GoogleSearch({'q':query
                  ,'api_key':'c1cc10ba909b3eda11d791d2fe5a6a4f32a3518830447dc68f634f36b27af4a9'
                  ,'device':'desktop'
                  ,'gl':country_code}).get_json()
            if perfectdict.has_key(json_results,'error'):
                return json_results
            for sr in json_results['organic_results']:
                results["search_results"][sr['position']] = {'title':sr['title'],'link':sr['link'],'snippet':perfectdict.key_ret(sr,'snippet')}
            if perfectdict.has_key(json_results,'related_questions'):
                for rq in json_results['related_questions']:
                    results["related_questions"][rq['question']] = {'title':rq['title'],'snippet':rq['snippet'],'date':perfectdict.key_ret(rq,'date'),'link':rq['link']}
            if perfectdict.has_key(json_results,'discussions_and_forums'):
                for daf in range(len(json_results['discussions_and_forums'])):
                    results['discussions_and_forums'][daf] = {'title':json_results['discussions_and_forums'][daf]['title'],'link':json_results['discussions_and_forums'][daf]['link'],'date':perfectdict.key_ret(json_results['discussions_and_forums'][daf],'date')}
            return results
        except Exception as e:
            if _debug:
                print("Error in google: ",e)
            return {"Error":"Failed","Further_Details_DEV":e}
    elif search_engine.lower() == 'bing':
        results = {"query":query,"search_engine":search_engine,"search_results":{}}
        try:
            json_results = serpapi.BingSearch({'q':query
                  ,'api_key':'c1cc10ba909b3eda11d791d2fe5a6a4f32a3518830447dc68f634f36b27af4a9'
                  ,'device':'desktop'
                  ,'cc':country_code}).get_json()
            if perfectdict.has_key(json_results,'error'):
                return json_results
            for sr in json_results['organic_results']:
                results["search_results"][sr['position']] = {'title':sr['title'],'link':sr['link'],'snippet':perfectdict.key_ret(sr,'snippet')}
            return results
        except Exception as e:
            if _debug:
                print("Error in bing: ",e)
            return {"Error":"Failed","Further_Details_DEV":e}
    else:
        return {"Error":"Failed","Further_Details_DEV":e}





ALL_AVAILABLE_TOOLS = [
    gen_func("run_py","Runs the given python code using 'exec' function",python_code=xtype(str,"The python code to run"))  ,
    gen_func("get_weather","Returns the current weather of input location",required=['location'],location=xtype(str,"the location to get the weather for, examples : 'Tehran','New York'"),unit=xtype(str,"The unit of temperature example: celcius",enum=['celcius','fahrenheit'])),
    gen_func("MessageBox","Shows a message box client-side to user using (pywin32.win32ui)",required=["message","title"],message=xtype(str,"The message shown in message box"),title=xtype(str,"The title shown for message box window"),style=xtype(int,"The style of the message box based on win32ui.MessageBox style policy range from 0 to 6. 0 being a basic message box with [ok] button and 1 being [ok,cancel] and so on...",enum=[0,1,2,3,4,5,6])),
    gen_func("OpenApp","Opens an application by searching the given name example: Notepad",appname=xtype(str,"The name of the application you want to open")) ,
    gen_func("ListInstalledAppsName","Returns a list of all applications installed on system"),
    gen_func("UninstallApp","Ask user for verification before uninstalling the given app",appname=xtype(str,"The name of the application you want to uninstall example:Pycharm")),
    gen_func("OpenUrl","simply opens a url in browser using webbrowser library",url=xtype(str,"the url that you want to open in browser")),
    gen_func("convert_path","Converts linux path to windows path, also used for using linux path",path=xtype(str,"the path you want to convert to supported path")),
    gen_func("OpenPath","simply opens a path in windows explorer, code:`explorer.exe {path}`",path=xtype(str,"The path to open in windows explorer")),
    gen_func("cython_inline","A function that will get an argument as input for cython.inline function inline compiler and returns a output",inline_text=xtype(str,"The inline string to be compiled and executed with cython.inline function.")),
    gen_func("run_cmd","Runs the given command line/shell code using subprocess and returns output",cmd_code=xtype(str,"The cmd code to run")),

]

def call_func(func_name:str,func_args:dict):
    return globals()[func_name](**func_args)




import json


# Formatation Section...
class TTArray:
    def __init__(self,items_type:type):
        self.items_type=items_type
    @property
    def t(self):
        return self.items_type

def xtype(type_,description:str,enum:list=None):
    if description == 'no_desc':
        if type_ == str and enum != None:
            return {"type":"string","enum":enum}
        elif type_ == int and enum != None:
            return {"type":"integer","enum":enum}
        elif type_ == float and enum != None:
            return {"type":"float","enum":enum}
        elif type_ == str:
            return {"type":"string"}
        elif type_ == int:
            return {"type":"integer"}
        elif type_ == float:
            return {"type":"float"}
        elif isinstance(type_,TTArray):
            return {"type":"array","items":type_.t}
        else:
            raise ValueError("Invalid value for xtype.argumentParser")
    if type_ == str and enum != None:
        return {"type":"string","enum":enum,"description":description}
    elif type_ == int and enum != None:
        return {"type":"integer","enum":enum,"description":description}
    elif type == float and enum != None:
        return {"type":"float","enum":enum,"description":description}
    elif type_ == str:
        return {"type":"string","description":description}
    elif type_ == int:
        return {"type":"integer","description":description}
    elif type_ == float:
        return {"type":"float","description":description}
    elif isinstance(type_,TTArray):
        return {"type":"array","description":description,"items":type_.t}
    else:
        raise ValueError("Invalid value for xtype.argumentParser")

def gen_func(func_name:str,func_desc:str,required=all,**kwargs): # example -> gen_func("insert_members",required=["mname","mbudget"],mname=xtype(str,"member name"),mbudget=xtype(int,"member budget in us dollars",enum=[100,200,300,400]),keywords=xtype(TTArray(xtype(str,"no_desc")),"keyword"))
    properties = {}
    for key in kwargs:
        properties[key] = kwargs[key]
    
    return {
        "type": "function",
        "function": {
            "name": func_name,
            "description":func_desc,
            "parameters": {
                "type": "object",
                "properties": properties,
                "required": required if required != all else list(kwargs.keys()),
                "additionalProperties": False,
            },
        },
    }
def gen_func_response(tool_call_id:str,func_name:str,func_output:json): # out.choices[0].message.tool_calls[n].id
    return {
        "tool_call_id":tool_call_id,
        "role":"tool",
        "name":func_name,
        "content":func_output if isinstance(func_output,str) else json.dumps(func_output)
    }
